/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ds485.h"

#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <netdb.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/select.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>

#include <digitalSTROM/ds485-socket-server.h>
#include <digitalSTROM/ds485-socket-interface.h>

#include "dsmapi.h"
#include "dsm.h"
#include "datamodel.h"
#include "bsp.h"
#include "mmi.h"
#include "bitset8.h"
#include "types.h"
#include "util.h"
#include "vdc-api-proto.h"

#define LOG(_loglevel, _message, ...) vdsm_report_extraLevel(_loglevel, LOG_NOTICE, _message, ##__VA_ARGS__)

Q_DEFINE_THIS_MODULE(ds485)

typedef struct dS485Tag {
    QActive super;
    dsuid_t dsmid;
    bool terminate;
    bool listenGlobal;
    const char* listenPort;
    pthread_t rxThread;
    struct dS485_server* netlibHandle;
} DS485;

/* Active object class -----------------------------------------------------*/

static QState DS485_initial (DS485 *me, QEvent const *e);
static QState DS485_wait_for_configuration(DS485 *me, QEvent const *e);
static QState DS485_top (DS485 *me, QEvent const *e);
static QState DS485_startup(DS485 *me, QEvent const *e);

static QState DS485_operational(DS485 *me, QEvent const *e);

static QState DS485_stopped(DS485 *me, QEvent const *e);

static void* rxThread(void* arg);

/* Local objects -----------------------------------------------------------*/
static DS485 l_ds485;   /* the single instance of the dS485 active object */
static QEvent const * l_ds485TxMessageQueueSto[128];    /* event queue storage */
static const QEvent l_startupEnableEvt = { DS485_RX_STARTUP_SIG, 0 };
static QEQueue l_ds485TxMessageQueue;   /* transmit message queue */


/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_DS485 = (QActive *)&l_ds485;      /* "opaque" AO pointer */

/*..........................................................................*/
void DS485_ctor(const char* listenPort, bool listenGlobal)
{
    DS485 *me = &l_ds485;
    QActive_ctor(&me->super, (QStateHandler)&DS485_initial);
    /* initialize message event transmit queue */
    QEQueue_init(&l_ds485TxMessageQueue, l_ds485TxMessageQueueSto, Q_DIM(l_ds485TxMessageQueueSto), "ds485TxMessage", &me->super);
    /* initialize time events */
    me->terminate = FALSE;
    me->listenGlobal = listenGlobal;
    me->listenPort = listenPort;
}

uint8_t DS485_GetTxMessageQueueMargin(void)
{
    return QEQueue_getNFree(&l_ds485TxMessageQueue);
}

uint8_t DS485_GetDeferredTxMessageQueueMargin(void)
{
    return 0;
}

void DS485_EmitBusChange(const char *dsuid, uint8_t flags)
{
    dsuid_t vdcuid;
    dsuid_from_string(dsuid, &vdcuid);

    DS485ReqBusChangeEvt * msgEvt = Q_NEW(DS485ReqBusChangeEvt, DS485_RX_JOIN_SIG);
    msgEvt->dsuid = vdcuid;
    msgEvt->flags = flags;
    QActive_postFIFO(AO_DS485, (QEvent *)msgEvt);
}

/**
 * Create DS485 control message
 */
// not used in vdsm
DS485MessageEvt * DS485_CreateControlMessage(const dsuid_t *dest, uint8_t src, DS485CommandType type, DS485CommandId cmd, uint8_t payloadLength)
{
    uint8_t* message;
    Q_ASSERT(payloadLength <= DS485_SHORT_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485MessageEvt * msgEvt = Q_NEW(DS485MessageEvt, DS485_TX_MSG_SIG);
    /* build header */
    message = msgEvt->message;
    message[DS485_START_OFFSET] = DS485_START_BYTE;
    message[DS485_DEST_OFFSET] = DS485_TYPE_CONTROL;
    message[DS485_SRC_OFFSET] = src | type;
    message[DS485_CL_OFFSET] = (cmd << 4) | (payloadLength & 0xF);
    msgEvt->destinationAddr = *dest;
    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
    return msgEvt;
}

/**
 * Create DS485 data message
 */
DS485MessageEvt * DS485_CreateDataMessageFromId(const dsuid_t *destDsuid, uint16_t sourceId, uint8_t src, DS485CommandType type, uint8_t payloadLength)
{
    dsuid_t source_dsuid;
    if (sourceId == 0) {
        /* use vdsm dsuid as source */
        source_dsuid = Datamodel_GetDsmid();
    } else if (!Datamodel_GetVdcDsuidFromDeviceId(sourceId, &source_dsuid)) {
        vdsm_report(LOG_DEBUG, "can not convert source id %d to dsuid_t\n", sourceId);
        source_dsuid = DSUID_NULL;
    }
    return DS485_CreateDataMessageFromSrcDestDsuid(destDsuid, &source_dsuid, src, type, payloadLength);
}

DS485MessageEvt * DS485_CreateDataMessageFromSrcDestDsuid(const dsuid_t *destDsuid, const dsuid_t *sourceDsuid, uint8_t src, DS485CommandType type, uint8_t payloadLength)
{
    uint8_t* message;
    Q_ASSERT(payloadLength <= DS485_LONG_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485MessageEvt * msgEvt;
    if (payloadLength <= DS485_SHORT_PAYLOAD_SIZE) {
        msgEvt = (DS485MessageEvt *) Q_NEW(DS485ShortMessageEvt, DS485_TX_MSG_SIG);
    } else {
        msgEvt = Q_NEW(DS485MessageEvt, DS485_TX_MSG_SIG);
    }
    /* build header */
    message = msgEvt->message;
    message[DS485_START_OFFSET] = DS485_START_BYTE;
    message[DS485_DEST_OFFSET] = DS485_TYPE_DATA;
    message[DS485_SRC_OFFSET] = src | type;
    message[DS485_CL_OFFSET] = payloadLength;
    msgEvt->destinationAddr = *destDsuid;
    msgEvt->sourceAddr = *sourceDsuid;

    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
    return msgEvt;
}

/**
 * Create DS485 long data message
 */
DS485NetlibPacketMessageEvt * DS485_CreateLongDataMessage(const dsuid_t *destDsuid,
                                                          const dsuid_t *sourceDsuid,
                                                          uint8_t src,
                                                          DS485CommandType type,
                                                          uint16_t payloadLength)
{
    uint8_t* message;
    Q_ASSERT(payloadLength <= DS485N_PACKET_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485NetlibPacketMessageEvt *msgEvt = Q_NEW(DS485NetlibPacketMessageEvt, DS485_TX_LONG_MSG_SIG);
    /* build header */
    message = msgEvt->message;
    message[DS485N_PACKET_START_OFFSET] = DS485_START_BYTE;
    message[DS485N_PACKET_DEST_OFFSET] = DS485_TYPE_DATA;
    message[DS485N_PACKET_SRC_OFFSET] = src | type;
    message[DS485N_PACKET_CL_OFFSET_LOW] = ((payloadLength >> 8) & 0xff);
    message[DS485N_PACKET_CL_OFFSET_HIGH] = (payloadLength & 0xff);
    unaligned_memcpy((void *)&msgEvt->sourceAddr, (void *)sourceDsuid, sizeof(dsuid_t));
    unaligned_memcpy((void *)&msgEvt->destinationAddr, (void *)destDsuid, sizeof(dsuid_t));
    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
    return msgEvt;
}

/**
 * Create DS485 data message
 */
DS485MessageEvt * DS485_CreateDataMessageToDsm(uint8_t broadcast,
                                               const dsuid_t *srcId,
                                               const dsuid_t *destinationId,
                                               DS485CommandType type,
                                               uint8_t payloadLength)
{
    uint8_t* message;
    Q_ASSERT(payloadLength <= DS485_LONG_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485MessageEvt * msgEvt;
        msgEvt = Q_NEW(DS485MessageEvt, DS485_RX_DATA_SIG);
    /* build header */
    message = msgEvt->message;
    message[DS485_START_OFFSET] = DS485_START_BYTE;
    message[DS485_DEST_OFFSET] = broadcast | DS485_TYPE_DATA;
    message[DS485_SRC_OFFSET] = type;
    message[DS485_CL_OFFSET] = payloadLength;
    unaligned_memcpy((void *)&msgEvt->sourceAddr, (void *)srcId, sizeof(dsuid_t));
    unaligned_memcpy((void *)&msgEvt->destinationAddr, (void *)destinationId, sizeof(dsuid_t));
    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
    return msgEvt;
}

DS485NetlibPacketMessageEvt * DS485_CreateLongDataMessageToDsm(uint8_t dest, const dsuid_t *srcDsuid, const dsuid_t *destDsuid, DS485CommandType type, uint16_t payloadLength)
{
    uint8_t* message;
    Q_ASSERT(payloadLength <= DS485N_PACKET_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485NetlibPacketMessageEvt * msgEvt;
        msgEvt = Q_NEW(DS485NetlibPacketMessageEvt, DS485_RX_LONG_DATA_SIG);
    /* build header */
    message = msgEvt->message;
    message[DS485N_PACKET_START_OFFSET] = DS485_START_BYTE;
    message[DS485N_PACKET_DEST_OFFSET] = dest | DS485_TYPE_DATA;
    message[DS485N_PACKET_SRC_OFFSET] = type;
    message[DS485N_PACKET_CL_OFFSET_LOW] = ((payloadLength >> 8) & 0xff);
    message[DS485N_PACKET_CL_OFFSET_HIGH] = (payloadLength & 0xff);
    unaligned_memcpy((void *)&msgEvt->sourceAddr, (void *)srcDsuid, sizeof(dsuid_t));
    unaligned_memcpy((void *)&msgEvt->destinationAddr, (void *)destDsuid, sizeof(dsuid_t));
    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
    return msgEvt;
}

/**
 * Return next DS485 message to transmit or null
 */
DS485MessageEvt const * DS485_GetNextTxMessage(void)
{
    return (DS485MessageEvt const *)QEQueue_get(&l_ds485TxMessageQueue);
}

/*..........................................................................*/
static QState DS485_initial(DS485 *me, QEvent const *e)
{
    (void)e;        /* suppress the compiler warning about unused parameter */

    /* subscribe events */
    QActive_subscribe((QActive *)me, CONFIG_INIT_SIG);
    return Q_TRAN(&DS485_wait_for_configuration);
}

/*..........................................................................*/
static QState DS485_wait_for_configuration(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case CONFIG_INIT_SIG: {
        DatamodelConfigEvt const * config = (DatamodelConfigEvt const *)e;
        /* save dSM ID */
        me->dsmid = config->dsmid;
        /* subscribe events */
        QActive_subscribe((QActive *)me, SYSTEM_TERM_SIG);
        return Q_TRAN(&DS485_startup);
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState DS485_top(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case SYSTEM_TERM_SIG: {
        return Q_TRAN(&DS485_stopped);
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState DS485_startup(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* startup receiver thread */
        if (pthread_create(&me->rxThread, NULL, rxThread, me)) {
            LOG(LOG_ERR, "Could not create receiver thread.\n");
            /* TODO: die */
        }
        return Q_HANDLED();
    }
    case DS485_RX_STARTUP_SIG: {
        return Q_TRAN(&DS485_operational);
    }
//    case DS485_TX_MSG_SIG: {
//        /* defer transmit message event */
//        if (QEQueue_getNFree(&me->deferredTxMessageQueue) != 0) {
//            QActive_defer((QActive *)me, &me->deferredTxMessageQueue, e);
//        }
//        return Q_HANDLED();
//    }
    case Q_EXIT_SIG: {
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_operational(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DS485AddressEvt * addrEvt = Q_NEW(DS485AddressEvt, DS485_ADDRESS_SIG);
        addrEvt->deviceAddress = 0;
        QF_publish((QEvent *)addrEvt);
        return Q_HANDLED();
    }
    case DS485_TX_MSG_SIG: {
        DS485MessageEvt const * msgEvt = (DS485MessageEvt const *)e;
        ds485_container_t container;

        if (DS485_IsRequest(msgEvt)) {
            container.containerType = DS485_CONTAINER_REQUEST;
        } else if (DS485_IsResponse(msgEvt)) {
            container.containerType = DS485_CONTAINER_RESPONSE;
        } else {
            container.containerType = DS485_CONTAINER_EVENT;
        }

        container.destinationId = msgEvt->destinationAddr;
        container.sourceId = msgEvt->sourceAddr;

        container.length = DS485_GetPayloadLength(msgEvt) - 1;
        memcpy(container.raw_data, &msgEvt->message[DS485_DATA_OFFSET],
               container.length + 1);
        ds485_server_reply(me->netlibHandle, &container);
        return Q_HANDLED();
    }
    case DS485_TX_LONG_MSG_SIG: {
        DS485NetlibPacketMessageEvt const * msgEvt = (DS485NetlibPacketMessageEvt const *)e;
        ds485n_packet_t packet;

        if (DS485_IsRequest(msgEvt)) {
            packet.containerType = DS485_CONTAINER_REQUEST;
        } else if (DS485_IsResponse(msgEvt)) {
            packet.containerType = DS485_CONTAINER_RESPONSE;
        } else {
            packet.containerType = DS485_CONTAINER_EVENT;
        }

        packet.destinationId = msgEvt->destinationAddr;
        packet.sourceId      = msgEvt->sourceAddr;

        packet.length = DS485_GetNetlibPacketPayloadLength(msgEvt) - 1;
        memcpy(packet.raw_data, &msgEvt->message[DS485N_PACKET_DATA_OFFSET],
               packet.length + 1);

        ds485_server_reply_long(me->netlibHandle, &packet);
        return Q_HANDLED();
    }
    case DS485_RX_JOIN_SIG: {
        DS485ReqBusChangeEvt const * msgEvt = (DS485ReqBusChangeEvt const *)e;
        ds485_server_emit_bus_change(me->netlibHandle, msgEvt->dsuid, msgEvt->flags);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_stopped(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* wait till thread is terminated */
        int s;
        void *res;
        me->terminate = TRUE;
        ds485_server_abort(me->netlibHandle);
        s = pthread_join(me->rxThread, &res);
        if (s != 0) {
            LOG(LOG_ERR, "DS485_stopped: pthread_join returned error %d\n", s);
        }
        free(res);
        ds485_server_cleanup(me->netlibHandle);
        /* send DS485 stopped event */
        static const QEvent ds485StoppedEvt = { DS485_STOPPED_SIG, 0 };
        QActive_postFIFO(AO_MMI, &ds485StoppedEvt);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

int ds485_server_send_data(struct dS485_server *server,
                           const ds485_container_t* packet)
{
    (void) server;
    DS485MessageEvt *evt;
    uint8_t broadcast;
    DS485 *me = &l_ds485;

    if (!(dsuid_is_broadcast(&packet->destinationId) ||
          vdc_api_vdsm_holds_vdc(&packet->destinationId) ||
          dsuid_equal(&me->dsmid, &packet->destinationId))) {
        return 0;
    }

    broadcast = dsuid_is_broadcast(&packet->destinationId) ?  DS485_BROADCAST_ADDRESS : 0;
    evt = DS485_CreateDataMessageToDsm(broadcast,
                                       &packet->sourceId,
                                       &packet->destinationId,
                                       DS485_TYPE_REQ,
                                       packet->length + 1);

    unaligned_memcpy((void *)&evt->message[DS485_DATA_OFFSET],
                     (void *)packet->raw_data, packet->length + 1);
    QActive_postFIFO(AO_DsmApi, (QEvent *)evt);
    return 0;
}

int ds485_server_send_long_data(struct dS485_server *server,
                                const ds485n_packet_t* packet)
{
    (void) server;
    DS485NetlibPacketMessageEvt *evt;
    uint8_t dest;
    DS485 *me = &l_ds485;

    if (dsuid_is_broadcast(&packet->destinationId) ||
        dsuid_equal(&me->dsmid, &packet->destinationId)) {
        return 0;
    }

    dest = dsuid_is_broadcast(&packet->destinationId) ?  DS485_BROADCAST_ADDRESS : 0;
    evt = DS485_CreateLongDataMessageToDsm(dest,
                                           &packet->sourceId,
                                           &packet->destinationId,
                                           DS485_TYPE_REQ,
                                           packet->length + 1);

    unaligned_memcpy((void *)&evt->message[DS485N_PACKET_DATA_OFFSET],
                     (void *)packet->raw_data, packet->length + 1);

    QActive_postFIFO(AO_DsmApi, (QEvent *)evt);
    return 0;
}

int ds485_server_get_bus_members(struct dS485_server *server,
                                 dsuid_t *devices, int length)
{
    (void) server;
    int count = vdc_api_get_vdc_dsuid(devices, length);

    if (count < length) {
        DS485 *me = &l_ds485;
        memcpy(&devices[count], &(me->dsmid), sizeof(dsuid_t));
        ++count;
    }
    return count;
}

dsuid_t ds485_server_get_dsuid(struct dS485_server *server)
{
    (void) server;
    DS485 *me = &l_ds485;
    return me->dsmid;
}

int ds485_server_restart_bus(struct dS485_server *server)
{
    (void) server;
    return 0;
}

int ds485_server_set_packetfilter(struct dS485_server *server, uint8_t rxfilter)
{
    (void) server;
    (void) rxfilter;
    return 0;
}

struct dS485_server_callbacks netlib_cb =
{
    .receive_data = ds485_server_send_data,
    .receive_long_data = ds485_server_send_long_data,
    .get_bus_members = ds485_server_get_bus_members,
    .restart_bus = ds485_server_restart_bus,
    .get_dsuid = ds485_server_get_dsuid,
    .set_packetfilter = ds485_server_set_packetfilter,
};

static void* rxThread(void* arg)
{
    DS485 *me = (DS485*)arg;

    me->netlibHandle = ds485_server_init(&netlib_cb, false);
    if (me->netlibHandle == NULL) {
        LOG(LOG_ERR, "Could not allocate netlib handle.\n");
        return NULL;
    }
    QActive_postFIFO((QActive*)me, &l_startupEnableEvt);
    /* call netlib server listen function */
    ds485_server_listen(me->netlibHandle, strtol(me->listenPort, NULL, 10), me->listenGlobal);
    return NULL;
}
